use serde::{Deserialize, Serialize};
use std::fs::File;
use std;
use std::io::{Read, Write};
use toml;
use xdg;
use xdg::BaseDirectories;

#[derive(Debug)]
pub enum ConfigError {
    IOError(std::io::Error),
    TomlDeError(toml::de::Error),
    TomlSeError(toml::ser::Error),
}

impl From<std::io::Error> for ConfigError {
    fn from(e: std::io::Error) -> Self {
        ConfigError::IOError(e)
    }
}

impl From<toml::de::Error> for ConfigError {
    fn from(e: toml::de::Error) -> Self {
        ConfigError::TomlDeError(e)
    }
}

impl From<toml::ser::Error> for ConfigError {
    fn from(e: toml::ser::Error) -> Self {
        ConfigError::TomlSeError(e)
    }
}

pub struct Config {
    pub xdg_dirs: BaseDirectories,
}

impl Config {
    pub fn new() -> Config {
        let xdg_dirs = xdg::BaseDirectories::with_prefix("gog-sync").unwrap();
        Config { xdg_dirs: xdg_dirs }
    }

    pub fn load<T>(&self, name: &str) -> Result<T, ConfigError>
        where T: Deserialize
    {
        let config_path = self.xdg_dirs.place_config_file(name)?;
        let mut config_file = File::open(config_path)?;

        let mut config_contents = String::new();
        config_file.read_to_string(&mut config_contents)?;

        match toml::from_str(config_contents.as_str()) {
            Ok(value) => Ok(value),
            Err(error) => Err(ConfigError::TomlDeError(error)),
        }
    }

    pub fn save<T>(&self, name: &str, content: &T) -> Result<(), ConfigError>
        where T: Serialize
    {
        let config_path = self.xdg_dirs.place_config_file(name)?;
        let content_toml = toml::to_string(content)?;

        let mut config_file = File::create(&config_path)?;
        match config_file.write_all(content_toml.as_bytes()) {
            Ok(_) => Ok(()),
            Err(error) => Err(ConfigError::IOError(error)),
        }
    }
}
