use config::{Config, ConfigError};
use http::{Http, HttpError};
use models;
use models::{Token, Game, Installer};
use serde_json;
use serde_json::Value;
use std::fs;
use std::io;
use std::io::Write;
use std::path::Path;

#[derive(Debug)]
pub enum GogError {
    Error(&'static str),
    ConfigError(ConfigError),
    HttpError(HttpError),
    SerdeError(serde_json::Error),
    IOError(io::Error),
}

impl From<ConfigError> for GogError {
    fn from(e: ConfigError) -> Self {
        GogError::ConfigError(e)
    }
}

impl From<HttpError> for GogError {
    fn from(e: HttpError) -> Self {
        GogError::HttpError(e)
    }
}

impl From<serde_json::Error> for GogError {
    fn from(e: serde_json::Error) -> Self {
        GogError::SerdeError(e)
    }
}

impl From<io::Error> for GogError {
    fn from(e: io::Error) -> Self {
        GogError::IOError(e)
    }
}

pub struct Gog<'a> {
    client_id: String,
    client_secret: String,
    redirect_uri: String,
    games_uri: String,
    http_client: &'a mut Http,
}

impl<'a> Gog<'a> {
    pub fn new(http_client: &'a mut Http) -> Gog<'a> {
        Gog {
            client_id: String::from("46899977096215655"),
            client_secret: String::from("9d85c43b1482497dbbce61f6e4aa173a433796eeae2ca8c5f6129f2dc4de46d9"),
            redirect_uri: String::from("https://embed.gog.com/on_login_success?origin=client"),
            games_uri: String::from("https://embed.gog.com/user/data/games"),
            http_client: http_client,
        }
    }

    pub fn login(&mut self) -> Result<(), GogError> {
        let config = Config::new();
        let mut token: Token = match config.load("token.toml") {
            Ok(value) => value,
            Err(_) => {
                let code = self.get_code()?;
                self.get_token(code.as_str())?
            }
        };

        if token.is_expired() {
            token = self.refresh_token(token.refresh_token.as_str())?;
        }

        config.save("token.toml", &token)?;

        let auth_header = format!("Authorization: Bearer {token}", token = token.access_token);
        self.http_client.add_header(auth_header.as_str())?;

        Ok(())
    }

    pub fn sync(&mut self, storage_path: &str) -> Result<(), GogError> {
        let game_ids = self.get_game_ids()?;

        for game_id in game_ids {
            let game = self.get_game(game_id)?;
            let game_hash = models::get_hash(&game);

            let game_root = Path::new(storage_path).join(&game.title);
            fs::create_dir_all(&game_root)?;

            for installer in game.installers {
                let installer_hash = models::get_hash(&installer);

                let installer_uri = format!("https://embed.gog.com{}", installer.manual_url);

                info!("downloading {} for {}...",
                      &installer.manual_url,
                      &game.title);
                self.http_client.download(installer_uri.as_str(), &game_root)?;
            }

            for extra in game.extras {
                let extra_hash = models::get_hash(&extra);

                let extra_uri = format!("https://embed.gog.com{}", extra.manual_url);

                info!("downloading {} for {}...", &extra.name, &game.title);
                self.http_client.download(extra_uri.as_str(), &game_root)?;
            }
        }

        Ok(())
    }

    fn get_code(&self) -> Result<String, GogError> {
        let auth_uri = self.auth_uri(self.client_id.as_str(), self.redirect_uri.as_str());

        println!("{}", auth_uri);

        let mut code = String::new();

        print!("Code: ");
        io::stdout().flush()?;

        io::stdin()
            .read_line(&mut code)
            .expect("Failed to read line");

        Ok(code)
    }

    fn get_token(&mut self, code: &str) -> Result<Token, GogError> {
        let token_uri = self.token_uri(self.client_id.as_str(),
                                       self.client_secret.as_str(),
                                       code,
                                       self.redirect_uri.as_str());

        let token_response = match self.http_client.get(token_uri.as_str()) {
            Ok(value) => value,
            Err(error) => return Err(GogError::HttpError(error)),
        };

        match serde_json::from_str(&token_response) {
            Ok(value) => Ok(value),
            Err(error) => Err(GogError::SerdeError(error)),
        }
    }

    fn refresh_token(&mut self, refresh_token: &str) -> Result<Token, GogError> {
        let token_refresh_uri = self.token_refresh_uri(self.client_id.as_str(),
                                                       self.client_secret.as_str(),
                                                       refresh_token);

        let token_response = match self.http_client.get(token_refresh_uri.as_str()) {
            Ok(value) => value,
            Err(error) => return Err(GogError::HttpError(error)),
        };

        match serde_json::from_str(&token_response) {
            Ok(value) => Ok(value),
            Err(error) => Err(GogError::SerdeError(error)),
        }
    }

    fn get_game_ids(&mut self) -> Result<Vec<u64>, GogError> {
        let response = self.http_client.get(self.games_uri.as_str())?;
        let game_ids_raw: Value = serde_json::from_str(response.as_str())?;
        let game_ids_serde = &game_ids_raw["owned"];

        let mut game_ids: Vec<u64> = Vec::new();

        if !game_ids_serde.is_array() {
            return Err(GogError::Error("Error parsing game ids."));
        }

        for game_id in game_ids_serde.as_array().unwrap() {
            let game_id_parsed = game_id.as_u64().unwrap_or(0);

            if game_id_parsed == 0 {
                error!("Cant parse game id {}", game_id);
                continue;
            }

            game_ids.push(game_id_parsed);
        }

        return Ok(game_ids);
    }

    fn get_game(&mut self, game_id: u64) -> Result<Game, GogError> {
        let game_uri = self.game_uri(game_id);

        let response = self.http_client.get(game_uri.as_str())?;

        let mut game: Game = serde_json::from_str(&response)?;

        let game_raw: Value = serde_json::from_str(response.as_str())?;
        let downloads = &game_raw["downloads"];

        for languages in downloads.as_array() {
            for language in languages {
                if !language.is_array() || language.as_array().unwrap().len() < 2 {
                    error!("Skipping a language for {}", game.title);
                    continue;
                }

                let installer_language = match language[0].as_str() {
                    Some(value) => value,
                    None => "",
                };

                if installer_language == "" {
                    error!("Skipping a language for {}", game.title);
                    continue;
                }

                for systems in language[1].as_object() {
                    for system in systems.keys() {
                        for real_downloads in systems.get(system) {
                            for real_download in real_downloads.as_array() {
                                let download = &real_download[0];

                                if !download.is_object() ||
                                   !download.as_object().unwrap().contains_key("manualUrl") {
                                    error!("Skipping an installer for {}", game.title);
                                    continue;
                                }

                                let installer = Installer {
                                    manual_url: String::from(download["manualUrl"]
                                        .as_str()
                                        .unwrap()),
                                    version: String::from(download["version"]
                                        .as_str()
                                        .unwrap_or("")),
                                    os: system.clone(),
                                    language: String::from(installer_language),
                                };

                                game.installers.push(installer);
                            }
                        }
                    }
                }
            }
        }

        Ok(game)
    }

    fn auth_uri(&self, client_id: &str, redirect_uri: &str) -> String {
        format!("https://auth.gog.\
                 com/auth?client_id={client_id}&redirect_uri={redirect_uri}&response_type=code&layout=client2",
                client_id = client_id,
                redirect_uri = redirect_uri)
    }

    fn token_uri(&self,
                 client_id: &str,
                 client_secret: &str,
                 code: &str,
                 redirect_uri: &str)
                 -> String {
        format!("https://auth.gog.\
                 com/token?client_id={client_id}&client_secret={client_secret}&grant_type=authorization_code&code={code}&redirect_uri={redirect_uri}",
                client_id = client_id,
                client_secret = client_secret,
                code = code.trim(),
                redirect_uri = redirect_uri)
    }

    fn token_refresh_uri(&self,
                         client_id: &str,
                         client_secret: &str,
                         refresh_token: &str)
                         -> String {
        format!("https://auth.gog.\
                 com/token?client_id={client_id}&client_secret={client_secret}&grant_type=refresh_token&refresh_token={refresh_token}",
                client_id = client_id,
                client_secret = client_secret,
                refresh_token = refresh_token)
    }

    fn game_uri(&self, game_id: u64) -> String {
        format!("https://embed.gog.com/account/gameDetails/{game_id}.json",
                game_id = game_id)
    }
}
